/*
   Copyright 2008-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

package com.notuvy.util;

import org.apache.log4j.Logger;

/**
 * A utility to perform periodic task which are keyed off a
 * given amount of idle time.  It creates its own (daemon)
 * thread to perform the action.  It is created with the delay
 * period.  It will only trigger if the UI activity idleness is
 * greater than this given period.  It can also be triggered
 * manually, and the action will also take place in the same
 * thread.
 *
 * User: murali
 * Date: Aug 29, 2008
 * Time: 5:20:10 PM
 */
public class IdleMonitor implements Runnable, ThreadControllable {

    //------------------------------------------------------------
    //- Class Variables

    private static final Logger LOG = Logger.getLogger(IdleMonitor.class);

    //------------------------------------------------------------
    //- Class Functions

    //------------------------------------------------------------
    //- Instance Variables

    private final long fFullSleepPeriod;
    private final Runnable fRunnable;
    private final Thread fThread;
    private final ThreadControl fControl = new ThreadControl();

    private boolean fManual = false;
    private long fLastManualTriggerTime;

    //------------------------------------------------------------
    //- Constructors

    public IdleMonitor(String pThreadName, long pFullSleepPeriod, Runnable pRunnable) {
        fFullSleepPeriod = pFullSleepPeriod;
        fRunnable = pRunnable;
        fThread = new Thread(this, pThreadName);

        IdleMonitorEventQueue.register();
        // LOG.debug("First scheduled for: " + new Date(WallClock.currentTimeMillis() + pFullSleepPeriod));

        getThread().setDaemon(true);
        getThread().start();
    }

    //------------------------------------------------------------
    //- Getters

    public long getFullSleepPeriod() {
        return fFullSleepPeriod;
    }

    private Runnable getRunnable() {
        return fRunnable;
    }

    private boolean isManual() {
        return fManual;
    }

    private long getLastManualTriggerTime() {
        return fLastManualTriggerTime;
    }

    private Thread getThread() {
        return fThread;
    }

    public ThreadControl getControl() {
        return fControl;
    }


    //------------------------------------------------------------
    //- Settors

    private void setManual(boolean pManual) {
        fManual = pManual;
    }

    private void setLastManualTriggerTime(long pLastManualTriggerTime) {
        fLastManualTriggerTime = pLastManualTriggerTime;
    }

    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    private void executeRunnable() {
        if (getControl().isAlive()) {
            try {
                getControl().setRunning(true);
                getRunnable().run();

                IdleMonitorEventQueue.touch();  // After the action is taken, mark activity.
            } catch (Throwable thr) {
                LOG.error("Fatal error with the scheduled task.", thr);
            } finally {
                getControl().setRunning(false);
            }
        }
    }

    private long handleManual() {
        long nextSleepPeriod;//  LOG.debug("Manual trigger");
        setManual(false);  // Reset.
        setLastManualTriggerTime(WallClock.currentTimeMillis());
        executeRunnable();
        nextSleepPeriod = getFullSleepPeriod();
        return nextSleepPeriod;
    }

    private long handleAutomatic() {
        long nextSleepPeriod;
        long lastActivity = Math.max(IdleMonitorEventQueue.getLastActivity(), getLastManualTriggerTime());
        long idlePeriod = WallClock.currentTimeMillis() - lastActivity;
        if (idlePeriod >= getFullSleepPeriod()) {
            // If idle longer than the full sleep period, then okay to do the action.
            // The next check is after the full time.
            executeRunnable();
            nextSleepPeriod = getFullSleepPeriod();
        } else {
            // Otherwise (idle was less than the full time), sleep for the difference.
            // This way, if there still was no activity, then we wake the full time since
            // the last activity.
            nextSleepPeriod = getFullSleepPeriod() - idlePeriod;
        }
        return nextSleepPeriod;
    }

    //------------------------------------------------------------
    //- Public Interface Functions

    public void manualTrigger() {
        setManual(true);
        getThread().interrupt();
    }

    //------------------------------------------------------------
    //- Class Interface Functions

    public void run() {
        long nextSleepPeriod = getFullSleepPeriod();
        //noinspection InfiniteLoopStatement
        do {
            try {
                Thread.sleep(nextSleepPeriod);
            } catch (InterruptedException ie) {
                // Do nothing.  A manual trigger would have interrupted.
            }

            nextSleepPeriod = isManual() ? handleManual() : handleAutomatic();

            // LOG.debug("Rescheduled for: " + new Date(WallClock.currentTimeMillis() + nextSleepPeriod));
        } while (getControl().isAlive());
    }

    //------------------------------------------------------------
    //- Inner Classes


    //------------------------------------------------------------
    //- Main

}
