package com.notuvy.util;

import org.apache.log4j.Logger;

/**
 * Flags which control the operation of a worker thread.
 * There is no actual functionality here.  Rather, the implementation of the worker
 * thread should properly use these flags to control its operation.
 *
 * @author murali
 */
public class ThreadControl {

    //------------------------------------------------------------
    //- Class Variables

    protected static final Logger LOG = Logger.getLogger("com.notuvy.util");


    //------------------------------------------------------------
    //- Class Functions


    //------------------------------------------------------------
    //- Instance Variables

    private boolean fAlive = true;
    private boolean fEnabled = true;
    private boolean fRunning = false;


    //------------------------------------------------------------
    //- Constructors

    public boolean isAlive() {
        return fAlive;
    }

    public boolean isEnabled() {
        return fEnabled;
    }

    /**
     * Is the thread engaged in actual work currently.
     *
     * @return  boolean indicating that work is currently being performed.
     */
    public boolean isRunning() {
        return fRunning;
    }


    //------------------------------------------------------------
    //- Gettors

    /**
     * Called by the worker thread framework when actual work is started or finished.
     *
     * @param pRunning Indication whether actual work is currently being performed.
     */
    protected void setRunning(boolean pRunning) {
        fRunning = pRunning;
    }


    //------------------------------------------------------------
    //- Settors


    //------------------------------------------------------------
    //- Private/Protected Utility Functions


    //------------------------------------------------------------
    //- Public Interface Functions

    @Override
    public String toString() {
        return String.format("ThreadControl[alive=%s, enabled=%s, running=%s]", isAlive(), isEnabled(), isRunning());
    }

    /**
     * Terminate (permanently) the worker.
     *
     * Once the flag is marked "not alive", it can never again be marked "alive".
     */
    public void stop() {
        fAlive = false;
        fEnabled = false;
    }

    /**
     * Temporarily disable the worker.
     * New work is allowed to be queued, but nothing will be acted on until resumed.
     */
    public void suspend() {
        fEnabled = false;
    }

    /**
     * Enable the worker to process any queued work.
     */
    public void resume() {
        if (isAlive()) {
            fEnabled = true;
        }
    }

    /**
     * If the worker has been stopped, wait for all outstanding work to finish.
     * New work will not be taken on, so upon exit of this method we know that
     * the worker has been cleanly and fully shut down.
     */
    public void waitForQuiescence() {
        if (isAlive()) {
            LOG.warn("Quiescence is only relevant for stopped threads.");
        } else {
            while (isRunning()) {
                try {
                    Thread.sleep(500);
                } catch (InterruptedException ie) {
                    // Do nothing.
                }
            }
        }
    }


    //------------------------------------------------------------
    //- Class Interface Functions


    //------------------------------------------------------------
    //- Inner Classes


    //------------------------------------------------------------
    //- Main


}
