/*
   Copyright 2006-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

package com.notuvy.util;

import java.util.*;

import org.apache.log4j.Logger;

/**
 * A separate thread engine that executes a given Runnable.
 * The purpose of this is to avoid doing "new Thread()" an arbitrary
 * number of times.  Instead, a single thread is kept and it processes
 * each give Runnable.
 *
 * @author  murali
 */
public class ThreadWorker implements Runnable, ThreadControllable {

    //------------------------------------------------------------
    //- Class Variables

    protected static final Logger LOG = Logger.getLogger(ThreadWorker.class);


    //------------------------------------------------------------
    //- Class Functions


    //------------------------------------------------------------
    //- Instance Variables

    private final Thread fThread;
    private final LinkedList<Runnable> fRunnables = new LinkedList<Runnable>();
    private final ThreadControl fControl = new ThreadControl();

    //------------------------------------------------------------
    //- Constructors

    public ThreadWorker(String pName) {
        fThread = new Thread(this, pName);

        getThread().start();
    }


    //------------------------------------------------------------
    //- Accessors

    private Thread getThread() {
        return fThread;
    }

    private LinkedList<Runnable> getRunnables() {
        return fRunnables;
    }

    public ThreadControl getControl() {
        return fControl;
    }


    //------------------------------------------------------------
    //- Settors


    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    protected void process() {
        Runnable runnable = null;

        synchronized (this) {
            if (getRunnables().isEmpty()) {
                try {
                    wait();
                } catch (InterruptedException ie) {
                    LOG.debug("ThreadWorker was interrupted.", ie);
                }
            } else {
                runnable = getRunnables().removeFirst();
            }
        }

        if (runnable != null) {
            //noinspection SynchronizationOnLocalVariableOrMethodParameter
            synchronized (runnable) {
                try {
                    getControl().setRunning(true);
                    runnable.run();
                } catch (Throwable thr) {
                    LOG.error("Unhandled error in thread; aborting.", thr);
                } finally {
                    runnable.notify();  // for sync.
                    getControl().setRunning(false);
                }
            }
        }
    }


    //------------------------------------------------------------
    //- Public Interface Functions

    public void enqueueSync(Runnable pRunnable) {
        if (getControl().isAlive()) {
            enqueueAsync(pRunnable);

            //noinspection SynchronizationOnLocalVariableOrMethodParameter
            synchronized (pRunnable) {
                try {
                    pRunnable.wait();  // block until it is executed.
                } catch (InterruptedException ie) {
                    LOG.error("A ThreadWorker client was interrupted.", ie);
                }
            }
        } else {
            LOG.warn("Worker is shutting down.");
        }
    }


    public void enqueueAsync(Runnable pRunnable) {
        if (getControl().isAlive()) {
            synchronized (this) {
                getRunnables().add(pRunnable);
                notify();  // trigger the execution if waiting.
            }
        } else {
            LOG.warn("Worker is shutting down.");
        }
    }


    //------------------------------------------------------------
    //- Class Interface Functions

    public void run() {
        try {
            while (getControl().isAlive()) {
                process();
            }
        } catch(Throwable thr) {
            LOG.error("Problem in ThreadExecution; thread worker is now dead.", thr);
        }
    }


    //------------------------------------------------------------
    //- Inner Classes



    //------------------------------------------------------------
    //- Main



}
